package com.showmejava.jtoh;

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2000, Eric M. Burke (ericb@showmejava.com)
// All rights reserved.
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met:
//
// - Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
// 
// - Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// 
// - Neither name of the ShowMeJava.com nor the names of its contributors may 
//   be used to endorse or promote products derived from this software without
//   specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE. 
/////////////////////////////////////////////////////////////////////////////

import java.io.*;
import java.text.SimpleDateFormat;
import java.util.Date;

/**
 * Convert Java source code into HTML.
 *
 * @author Eric M. Burke
 * @version $Id: JToH.java,v 1.3 2000/04/11 03:11:00 ericb Exp $
 */
public class JToH implements Serializable {
    private static final int BUF_SIZE = 2000;

    // initialize to non-null values for JavaBeans support
    private String title = "";
    private String header = "";
    private String footer = "";
    private String styleSheetName = "jtoh.css";
    private boolean useLineNumbers = false;
            
    /**
     * Entry point when running as a standalone application.
     */
    public static void main(String[] args) throws IOException {
        if (args.length < 1) {
            System.err.println("usage: java com.showmejava.jtoh.JToH "
                    + "javafile [htmlfile]");
            System.exit(1);
        }
        String javafile = null;
        String htmlfile = null;
        javafile = args[0];
        if (args.length > 1) {
            htmlfile = args[1];
        }
        
        BufferedReader in = new BufferedReader(new FileReader(javafile));
        PrintWriter out = null;
        if (htmlfile == null) {
            out = new PrintWriter(new OutputStreamWriter(System.out));
        } else {
            out = new PrintWriter(new BufferedWriter(
                    new FileWriter(htmlfile)));
        }
        JToH translator = new JToH();

        // create a header
        SimpleDateFormat sdf = new SimpleDateFormat("dd MMM yyyy");
        StringBuffer header = new StringBuffer();
        header.append("<h2>").append(javafile).append("</h2>\n");
        header.append("<b><i>HTML Generated on ");
        header.append(sdf.format(new Date()));
        header.append(" by <a href='http://www.showmejava.com/jtoh/'>");
        header.append("JToH</a></i></b>\n");
        header.append("<hr>\n");

        translator.setTitle(javafile);
        translator.setUseLineNumbers(true);
        translator.setHeader(header.toString());
        translator.translateHTML(in, out);
    }

    /**
     * @param useLineNumbers if true, include line numbers in the output.
     */
    public void setUseLineNumbers(boolean useLineNumbers) {
        this.useLineNumbers = useLineNumbers;
    }

    /**
     * @return true if line numbering is enabled.
     */
    public boolean getUseLineNumbers() {
        return useLineNumbers;
    }

    /**
     * @param title the title for the generated HTML web page.
     */
    public void setTitle(String title) {
        this.title = title;
    }

    /**
     * @return the title for the generated HTML web page.
     */
    public String getTitle() {
        return title;
    }

    /**
     * @param styleSheetName alternate name for the cascading style sheet.
     * The default value is jtoh.css.
     */
    public void setStyleSheetName(String styleSheetName) {
        this.styleSheetName = styleSheetName;
    }

    /**
     * @return the name of the cascading style sheet.
     */
    public String getStyleSheetName() {
        return styleSheetName;
    }

    /**
     * @param optional HTML to insert just before the generated code.
     */
    public void setHeader(String header) {
        this.header = header;
    }

    /**
     * @return HTML to insert just before the generated code.
     */
    public String getHeader() {
        return header;
    }

    /**
     * @param footer optional HTML to insert just after the generated code.
     */
    public void setFooter(String footer) {
        this.footer = footer;
    }

    /**
     * @return HTML to insert just after the generated code.
     */
    public String getFooter() {
        return footer;
    }


    /**
     * Perform Java to HTML conversion.
     *
     * @param in the Java source code.
     * @param out the destination for the HTML.  If null, simply send output
     *            to System.out.
     * @exception IOException if a file cannot be read from or written to.
     */
    public void translateHTML(BufferedReader in, PrintWriter out) 
            throws IOException {
        if (out == null) {
            out = new PrintWriter(new OutputStreamWriter(System.out));
        }

        out.println("<html>");
        out.println("<head>");
        out.println("<meta name=\"GENERATOR\" content=\"JToH\">");
        out.print("<title>");
        out.print(title);
        out.println("</title>");
        out.print("<link href=\"");
        out.print(styleSheetName);
        out.println("\" rel=\"stylesheet\" type=\"text/css\">");
        out.println("</head>");
        out.println("<body>");

        out.println(header);
        translateBody(in, out);
        out.println(footer);
        
        out.println("</body>");
        out.println("</html>");
        out.close();
    }


    /**
     * Translate just the source code portion.  You can optionally call this
     * method if you just want the preformatted source code, without all
     * of the other HTML tags.  This is useful if you wish to embed source
     * code into another web page that some Servlet is generating.
     *
     * @param in the input source containing Java code.
     * @param out the destination for HTML.
     * @exception IOException if I/O fails.
     */
    public void translateBody(BufferedReader in, PrintWriter out)
            throws IOException {
        out.print("<pre>");
        translate(in, out);
        out.println("</pre>");
    }
    
    
    // do the actual source code analysis and conversion
    private void translate(BufferedReader in, PrintWriter out) 
            throws IOException {

        // The destination initializes to out.  If line numbers are enabled,
        // then first write to an alternate destination.
        PrintWriter dest = out;
        CharArrayWriter cawOut =null;
        if (useLineNumbers) {
            cawOut = new CharArrayWriter();
            dest = new PrintWriter(cawOut);
        }

        // read the input document into an array of characters
        CharArrayWriter cawIn = new CharArrayWriter();
        char[] cbufIn = new char[BUF_SIZE];
        int numCharsRead = in.read(cbufIn, 0, cbufIn.length);
        while (numCharsRead > -1) {
            cawIn.write(cbufIn, 0, numCharsRead);
            numCharsRead = in.read(cbufIn, 0, cbufIn.length);
        }
        in.close();
        char[] characters = cawIn.toCharArray();
        
        WhitespaceState whitespaceState = new WhitespaceState(characters);
        
        State[] states = new State[] {
            new SingleLineCommentState(characters),
            new MultiLineCommentState(characters),
            new QuoteState(characters),
            new AposState(characters),
            new KeywordState(characters)
        };

        int curPos = 0;
        while (curPos < characters.length) {
            State curState = null;
            for (int i=0; i<states.length; i++) {
                if (states[i].isStartPos(curPos)) {
                    curState = states[i];
                    curState.setStartPos(curPos);
                    break;
                }
            }

            // if none of the states match, use WhitespaceState
            if (curState == null) {
                curState = whitespaceState;
                curState.setStartPos(curPos);
            }
            
            curState.write(dest);
            curPos = curState.getEndPos();
        }

        // append line numbers if enabled
        if (useLineNumbers) {
            // five characters of padding supports files with up to 
            // 99999 lines of code.
            char[] padding = "     ".toCharArray();

            BufferedReader br = new BufferedReader(new StringReader(
                   cawOut.toString())); 
            int curLineNumber = 1;
            String curLine = br.readLine();
            while (curLine != null) {
                char[] curLineNumberAsChars = 
                        Integer.toString(curLineNumber).toCharArray();

                // this will first happen on files containing more than
                // 99999 lines of code, which is unlikely but conceivable.
                if (curLineNumberAsChars.length > padding.length) {
                    padding = ((new String(padding)) + " ").toCharArray();
                }

                System.arraycopy(curLineNumberAsChars, 0, padding,
                        padding.length-curLineNumberAsChars.length,
                        curLineNumberAsChars.length);

                out.println("<font class=\"lineNum\">" + new String(padding) 
                        + " </font> " + curLine);
                curLineNumber++;
                curLine = br.readLine();
            }
        }
    }
}
