package com.showmejava.jtoh;

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) 2000, Eric M. Burke (ericb@showmejava.com)
// All rights reserved.
// Redistribution and use in source and binary forms, with or without 
// modification, are permitted provided that the following conditions are met:
//
// - Redistributions of source code must retain the above copyright notice,
//   this list of conditions and the following disclaimer.
// 
// - Redistributions in binary form must reproduce the above copyright notice,
//   this list of conditions and the following disclaimer in the documentation
//   and/or other materials provided with the distribution.
// 
// - Neither name of the ShowMeJava.com nor the names of its contributors may 
//   be used to endorse or promote products derived from this software without
//   specific prior written permission. 
// 
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS 
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
// TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
// PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE
// LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
// CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
// SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
// INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
// CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
// ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE. 
/////////////////////////////////////////////////////////////////////////////


import java.io.*;

/**
 * Abstract base class for all other states.  Each state knows about
 * a specific pattern in a source code file, such as a language keyword,
 * a String, or a comment.
 *
 * @author Eric M. Burke
 * @version $Id: State.java,v 1.2 2000/04/01 05:20:13 ericb Exp $
 */
abstract class State {
    /**
     * Constant which indicates the end of file has been reached.
     */
    static final char EOF = Character.MIN_VALUE;

    private char[] characters;
    private int startPos;
    private int endPos;

    // the complete list of Java keywords
    private static final String[] KEYWORDS = {
        "abstract",
        "boolean",
        "break",
        "byte",
        "case",
        "catch",
        "char",
        "class",
        "const",
        "double",
        "else",
        "extends",
        "final",
        "finally",
        "float",
        "for",
        "goto",
        "if",
        "int",
        "interface",
        "long",
        "native",
        "new",
        "null",
        "package",
        "private",
        "protected",
        "static",
        "super",
        "switch",
        "synchronized",
        "this",
        "throw",
        "throws",
        "transient",
        "try",
        "continue",
        "default",
        "do",
        "implements",
        "import",
        "instanceof",
        "public",
        "return",
        "short",
        "void",
        "volatile",
        "while"
    };

    /**
     * Construct a new state.
     *
     * @param characters the array of characters to analyze.  The array
     * contains the entire input Java source file.
     */
    State(char[] characters) {
        this.characters = characters;
    }
    
    /**
     * @return the position of the first character that this state applies to.
     */
    int getStartPos() {
        return startPos;
    }
    
    /**
     * @return the position of the last character that this state applies to.
     */
    int getEndPos() {
        return endPos;
    }
    
    /**
     * @param startPos the position to check in the file.
     * @return true if the specified position is the beginning
     * of this state.
     */
    abstract boolean isStartPos(int startPos);
    
    
    /**
     * Set the starting position.  It is important that the caller
     * first calls isStartPos to verify that startPos truly is the
     * correct starting location for this state.  The end position will
     * be calculated as a side effect of calling this method.
     */
    void setStartPos(int startPos) {
        this.startPos = startPos;
        this.endPos = findEndPos(startPos);
    }
    
    /**
     * Write this state out to the specified print writer.
     *
     * @param pw the writer to write to.
     */
    void write(PrintWriter pw) {
        // let derived classes write the first part of this tag
        writeBeginTag(pw);
        
        // write the characters
        for (int i=startPos; i<endPos; i++) {
            char ch = getCharAt(i);
            writeChar(pw, ch);
        }
        
        writeEndTag(pw);
    }
    
    /**
     * Write a beginning HTML tag, if appropriate.
     *
     * @param pw the writer to write to.
     */
    protected abstract void writeBeginTag(PrintWriter pw);

    /**
     * Write an ending HTML tag, if appropriate.
     *
     * @param pw the writer to write to.
     */
    protected abstract void writeEndTag(PrintWriter pw);
    
    /**
     * @param startPos the starting position of this state.
     * @return the end position, never exceeding the end of the file.
     */
    protected abstract int findEndPos(int startPos);
    
    private void writeChar(PrintWriter pw, char ch) {
        switch (ch) {
        case '<':
            pw.print("&lt;");
            break;
        case '>':
            pw.print("&gt;");
            break;
        case '"':
            pw.print("&quot;");
            break;
        default:
            pw.print(ch);
        }
    }   


    protected final char getCharAt(int pos) {
        return (pos > -1 && pos < characters.length) ?  characters[pos] : EOF;
    }

    protected int getLength() {
        return characters.length;
    }


    // tbd: this may be horribly slow.  
    // tbd: Need to implement a much better algorithm
    public String getKeywordAtPosition(int pos) {
        for (int i=0; i<KEYWORDS.length; i++) {
            int curLen = KEYWORDS[i].length();
            
            // check for end of file
            if (pos+curLen > characters.length) {
                continue;
            }
            
            if (isJavaIdentifierPart(pos-1) 
                    || isJavaIdentifierPart(pos+curLen)) {
                continue;
            }
            
            String strAtCurPos = new String(characters, pos, curLen);
            if (!strAtCurPos.equals(KEYWORDS[i])) {
                continue;
            }
            
            return KEYWORDS[i];
        }
        return null;
    }
    
    private boolean isJavaIdentifierPart(int pos) {
        if (pos > -1 && pos < characters.length) {
            return Character.isJavaIdentifierPart(characters[pos]);
        }
        return false;
    }


}
